# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import unittest
from datetime import datetime

from opencensus.trace import execution_context, time_event
from opencensus.trace.span_context import SpanContext
from opencensus.trace.status import Status as OcStatus
from opencensus.trace.tracer import Tracer as OcTracer

from opentelemetry import trace
from opentelemetry.sdk.trace import ReadableSpan, TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
    InMemorySpanExporter,
)
from opentelemetry.sdk.trace.sampling import ALWAYS_ON
from opentelemetry.shim.opencensus import install_shim, uninstall_shim

_TIMESTAMP = datetime.fromisoformat("2023-01-01T00:00:00.000000")


class TestShimWithSdk(unittest.TestCase):
    def setUp(self):
        uninstall_shim()
        self.tracer_provider = TracerProvider(
            sampler=ALWAYS_ON, shutdown_on_exit=False
        )
        self.mem_exporter = InMemorySpanExporter()
        self.tracer_provider.add_span_processor(
            SimpleSpanProcessor(self.mem_exporter)
        )
        install_shim(self.tracer_provider)

    def tearDown(self):
        uninstall_shim()

    def test_start_span_interacts_with_context(self):
        oc_tracer = OcTracer()
        span = oc_tracer.start_span("foo")

        # Should have created a real OTel span in implicit context under the hood. OpenCensus
        # does not require another step to set the span in context.
        otel_span = trace.get_current_span()
        self.assertNotEqual(span.span_id, 0)
        self.assertEqual(span.span_id, otel_span.get_span_context().span_id)

        # This should end the span and remove it from context
        oc_tracer.end_span()
        self.assertIs(trace.get_current_span(), trace.INVALID_SPAN)

    def test_start_span_interacts_with_oc_context(self):
        oc_tracer = OcTracer()
        span = oc_tracer.start_span("foo")

        # Should have put the shim span in OC's implicit context under the hood. OpenCensus
        # does not require another step to set the span in context.
        self.assertIs(execution_context.get_current_span(), span)

        # This should end the span and remove it from context
        oc_tracer.end_span()
        self.assertIs(execution_context.get_current_span(), None)

    def test_context_manager_interacts_with_context(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("foo") as span:
            # Should have created a real OTel span in implicit context under the hood
            otel_span = trace.get_current_span()

            self.assertNotEqual(span.span_id, 0)
            self.assertEqual(
                span.span_id, otel_span.get_span_context().span_id
            )

        # The span should now be popped from context
        self.assertIs(trace.get_current_span(), trace.INVALID_SPAN)

    def test_context_manager_interacts_with_oc_context(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("foo") as span:
            # Should have placed the shim span in implicit context under the hood
            self.assertIs(execution_context.get_current_span(), span)

        # The span should now be popped from context
        self.assertIs(execution_context.get_current_span(), None)

    def test_exports_a_span(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("span1"):
            pass

        self.assertEqual(len(self.mem_exporter.get_finished_spans()), 1)

    def test_uses_tracers_span_context_when_no_parent_in_context(self):
        # the SpanContext passed to the Tracer will become the parent when there is no span
        # already set in the OTel context
        oc_tracer = OcTracer(
            span_context=SpanContext(
                trace_id="ace0216bab2b7ba249761dbb19c871b7",
                span_id="1fead89ecf242225",
            )
        )

        with oc_tracer.start_span("span1"):
            pass

        exported_span: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        parent = exported_span.parent
        self.assertIsNotNone(parent)
        self.assertEqual(
            trace.format_trace_id(parent.trace_id),
            "ace0216bab2b7ba249761dbb19c871b7",
        )
        self.assertEqual(
            trace.format_span_id(parent.span_id), "1fead89ecf242225"
        )

    def test_ignores_tracers_span_context_when_parent_already_in_context(self):
        # the SpanContext passed to the Tracer will be ignored since there is already a span
        # set in the OTel context
        oc_tracer = OcTracer(
            span_context=SpanContext(
                trace_id="ace0216bab2b7ba249761dbb19c871b7",
                span_id="1fead89ecf242225",
            )
        )
        otel_tracer = self.tracer_provider.get_tracer(__name__)

        with otel_tracer.start_as_current_span("some_parent"):
            with oc_tracer.start_span("span1"):
                pass

        oc_span: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        otel_parent: ReadableSpan = self.mem_exporter.get_finished_spans()[1]
        self.assertEqual(
            oc_span.parent,
            otel_parent.context,
        )

    def test_span_attributes(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("span1") as span:
            span.add_attribute("key1", "value1")
            span.add_attribute("key2", "value2")

        exported_span: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        self.assertDictEqual(
            dict(exported_span.attributes),
            {"key1": "value1", "key2": "value2"},
        )

    def test_span_annotations(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("span1") as span:
            span.add_annotation("description", key1="value1", key2="value2")

        exported_span: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        self.assertEqual(len(exported_span.events), 1)
        event = exported_span.events[0]
        self.assertEqual(event.name, "description")
        self.assertDictEqual(
            dict(event.attributes), {"key1": "value1", "key2": "value2"}
        )

    def test_span_message_event(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("span1") as span:
            span.add_message_event(
                time_event.MessageEvent(
                    _TIMESTAMP, "id_sent", time_event.Type.SENT, "20", "10"
                )
            )
            span.add_message_event(
                time_event.MessageEvent(
                    _TIMESTAMP,
                    "id_received",
                    time_event.Type.RECEIVED,
                    "20",
                    "10",
                )
            )
            span.add_message_event(
                time_event.MessageEvent(
                    _TIMESTAMP,
                    "id_unspecified",
                    None,
                    "20",
                    "10",
                )
            )

        exported_span: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        self.assertEqual(len(exported_span.events), 3)
        event1, event2, event3 = exported_span.events

        self.assertEqual(event1.name, "id_sent")
        self.assertDictEqual(
            dict(event1.attributes),
            {
                "message.event.size.compressed": "10",
                "message.event.size.uncompressed": "20",
                "message.event.type": "SENT",
            },
        )
        self.assertEqual(event2.name, "id_received")
        self.assertDictEqual(
            dict(event2.attributes),
            {
                "message.event.size.compressed": "10",
                "message.event.size.uncompressed": "20",
                "message.event.type": "RECEIVED",
            },
        )
        self.assertEqual(event3.name, "id_unspecified")
        self.assertDictEqual(
            dict(event3.attributes),
            {
                "message.event.size.compressed": "10",
                "message.event.size.uncompressed": "20",
                "message.event.type": "TYPE_UNSPECIFIED",
            },
        )

    def test_span_status(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("span_ok") as span:
            # OTel will log about the message being set on a not OK span
            with self.assertLogs(level=logging.WARNING) as rec:
                span.set_status(OcStatus(0, "message"))
            self.assertIn(
                "description should only be set when status_code is set to StatusCode.ERROR",
                rec.output[0],
            )

        with oc_tracer.start_span("span_exception") as span:
            span.set_status(
                OcStatus.from_exception(Exception("exception message"))
            )

        self.assertEqual(len(self.mem_exporter.get_finished_spans()), 2)
        ok_span: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        exc_span: ReadableSpan = self.mem_exporter.get_finished_spans()[1]

        self.assertTrue(ok_span.status.is_ok)
        # should be none even though we provided it because OTel drops the description when
        # status is not ERROR
        self.assertIsNone(ok_span.status.description)

        self.assertFalse(exc_span.status.is_ok)
        self.assertEqual(exc_span.status.description, "exception message")

    def assert_related(self, *, child: ReadableSpan, parent: ReadableSpan):
        self.assertEqual(
            child.parent.span_id, parent.get_span_context().span_id
        )

    def test_otel_sandwich(self):
        oc_tracer = OcTracer()
        otel_tracer = self.tracer_provider.get_tracer(__name__)
        with oc_tracer.start_span("opencensus_outer"):
            with otel_tracer.start_as_current_span("otel_middle"):
                with oc_tracer.start_span("opencensus_inner"):
                    pass

        self.assertEqual(len(self.mem_exporter.get_finished_spans()), 3)
        opencensus_inner: ReadableSpan = (
            self.mem_exporter.get_finished_spans()[0]
        )
        otel_middle: ReadableSpan = self.mem_exporter.get_finished_spans()[1]
        opencensus_outer: ReadableSpan = (
            self.mem_exporter.get_finished_spans()[2]
        )

        self.assertEqual(opencensus_outer.name, "opencensus_outer")
        self.assertEqual(otel_middle.name, "otel_middle")
        self.assertEqual(opencensus_inner.name, "opencensus_inner")

        self.assertIsNone(opencensus_outer.parent)
        self.assert_related(parent=opencensus_outer, child=otel_middle)
        self.assert_related(parent=otel_middle, child=opencensus_inner)

    def test_opencensus_sandwich(self):
        oc_tracer = OcTracer()
        otel_tracer = self.tracer_provider.get_tracer(__name__)
        with otel_tracer.start_as_current_span("otel_outer"):
            with oc_tracer.start_span("opencensus_middle"):
                with otel_tracer.start_as_current_span("otel_inner"):
                    pass

        self.assertEqual(len(self.mem_exporter.get_finished_spans()), 3)
        otel_inner: ReadableSpan = self.mem_exporter.get_finished_spans()[0]
        opencensus_middle: ReadableSpan = (
            self.mem_exporter.get_finished_spans()[1]
        )
        otel_outer: ReadableSpan = self.mem_exporter.get_finished_spans()[2]

        self.assertEqual(otel_outer.name, "otel_outer")
        self.assertEqual(opencensus_middle.name, "opencensus_middle")
        self.assertEqual(otel_inner.name, "otel_inner")

        self.assertIsNone(otel_outer.parent)
        self.assert_related(parent=otel_outer, child=opencensus_middle)
        self.assert_related(parent=opencensus_middle, child=otel_inner)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import wrapt
from opencensus.trace import execution_context
from opencensus.trace.blank_span import BlankSpan
from opencensus.trace.span_context import SpanContext
from opencensus.trace.tracers.base import Tracer as BaseTracer
from opencensus.trace.tracestate import Tracestate

from opentelemetry import context, trace
from opentelemetry.shim.opencensus._shim_span import ShimSpan

_logger = logging.getLogger(__name__)

_SHIM_SPAN_KEY = context.create_key("opencensus-shim-span-key")
_SAMPLED = trace.TraceFlags(trace.TraceFlags.SAMPLED)


def set_shim_span_in_context(
    span: ShimSpan, ctx: context.Context
) -> context.Context:
    return context.set_value(_SHIM_SPAN_KEY, span, ctx)


def get_shim_span_in_context() -> ShimSpan:
    return context.get_value(_SHIM_SPAN_KEY)


def set_oc_span_in_context(
    oc_span_context: SpanContext, ctx: context.Context
) -> context.Context:
    """Returns a new OTel context based on ctx with oc_span_context set as the current span"""

    # If no SpanContext is passed to the opencensus.trace.tracer.Tracer, it creates a new one
    # with a random trace ID and a None span ID to be the parent:
    # https://github.com/census-instrumentation/opencensus-python/blob/2e08df591b507612b3968be8c2538dedbf8fab37/opencensus/trace/tracer.py#L47.
    #
    # OpenTelemetry considers this an invalid SpanContext and will ignore it, so we can just
    # return early
    if oc_span_context.span_id is None:
        return ctx

    trace_id = int(oc_span_context.trace_id, 16)
    span_id = int(oc_span_context.span_id, 16)
    is_remote = oc_span_context.from_header
    trace_flags = (
        _SAMPLED if oc_span_context.trace_options.get_enabled() else None
    )
    trace_state = (
        trace.TraceState(tuple(oc_span_context.tracestate.items()))
        # OC SpanContext does not validate this type
        if isinstance(oc_span_context.tracestate, Tracestate)
        else None
    )

    return trace.set_span_in_context(
        trace.NonRecordingSpan(
            trace.SpanContext(
                trace_id=trace_id,
                span_id=span_id,
                is_remote=is_remote,
                trace_flags=trace_flags,
                trace_state=trace_state,
            )
        )
    )


# pylint: disable=abstract-method
class ShimTracer(wrapt.ObjectProxy):
    def __init__(
        self,
        wrapped: BaseTracer,
        *,
        oc_span_context: SpanContext,
        otel_tracer: trace.Tracer,
    ) -> None:
        super().__init__(wrapped)
        self._self_oc_span_context = oc_span_context
        self._self_otel_tracer = otel_tracer

    # For now, finish() is not implemented by the shim. It would require keeping a list of all
    # spans created so they can all be finished.
    # def finish(self):
    #    """End spans and send to reporter."""

    def span(self, name="span"):
        return self.start_span(name=name)

    def start_span(self, name="span"):
        parent_ctx = context.get_current()
        # If there is no current span in context, use the one provided to the OC Tracer at
        # creation time
        if trace.get_current_span(parent_ctx) is trace.INVALID_SPAN:
            parent_ctx = set_oc_span_in_context(
                self._self_oc_span_context, parent_ctx
            )

        span = self._self_otel_tracer.start_span(name, context=parent_ctx)
        shim_span = ShimSpan(
            BlankSpan(name=name, context_tracer=self),
            otel_span=span,
            shim_tracer=self,
        )

        ctx = trace.set_span_in_context(span)
        ctx = set_shim_span_in_context(shim_span, ctx)

        # OpenCensus's ContextTracer calls execution_context.set_current_span(span) which is
        # equivalent to the below. This can cause context to leak but is equivalent.
        # pylint: disable=protected-access
        shim_span._self_token = context.attach(ctx)
        # Also set it in OC's context, equivalent to
        # https://github.com/census-instrumentation/opencensus-python/blob/2e08df591b507612b3968be8c2538dedbf8fab37/opencensus/trace/tracers/context_tracer.py#L94
        execution_context.set_current_span(shim_span)
        return shim_span

    def end_span(self):
        """Finishes the current span in the context and restores the context from before the
        span was started.
        """
        span = self.current_span()
        if not span:
            _logger.warning("No active span, cannot do end_span.")
            return

        span.finish()

        # pylint: disable=protected-access
        context.detach(span._self_token)
        # Also reset the OC execution_context, equivalent to
        # https://github.com/census-instrumentation/opencensus-python/blob/2e08df591b507612b3968be8c2538dedbf8fab37/opencensus/trace/tracers/context_tracer.py#L114-L117
        execution_context.set_current_span(self.current_span())

    # pylint: disable=no-self-use
    def current_span(self):
        return get_shim_span_in_context()

    def add_attribute_to_current_span(self, attribute_key, attribute_value):
        self.current_span().add_attribute(attribute_key, attribute_value)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import lru_cache
from logging import getLogger
from typing import Optional

from opencensus.trace.span_context import SpanContext
from opencensus.trace.tracer import Tracer
from opencensus.trace.tracers.noop_tracer import NoopTracer

from opentelemetry import trace
from opentelemetry.shim.opencensus._shim_tracer import ShimTracer
from opentelemetry.shim.opencensus.version import __version__

_logger = getLogger(__name__)


def install_shim(
    tracer_provider: Optional[trace.TracerProvider] = None,
) -> None:
    otel_tracer = trace.get_tracer(
        "opentelemetry-opencensus-shim",
        __version__,
        tracer_provider=tracer_provider,
    )

    @lru_cache()
    def cached_shim_tracer(span_context: SpanContext) -> ShimTracer:
        return ShimTracer(
            NoopTracer(),
            oc_span_context=span_context,
            otel_tracer=otel_tracer,
        )

    def fget_tracer(self: Tracer) -> ShimTracer:
        # self.span_context is how instrumentations pass propagated context into OpenCensus e.g.
        # https://github.com/census-instrumentation/opencensus-python/blob/fd064f438c5e490d25b004ee2545be55d2e28679/contrib/opencensus-ext-flask/opencensus/ext/flask/flask_middleware.py#L147-L153
        return cached_shim_tracer(self.span_context)

    def fset_tracer(self, value) -> None:
        # ignore attempts to set the value
        pass

    # Tracer's constructor sets self.tracer to either a NoopTracer or ContextTracer depending
    # on sampler:
    # https://github.com/census-instrumentation/opencensus-python/blob/2e08df591b507612b3968be8c2538dedbf8fab37/opencensus/trace/tracer.py#L63.
    # We monkeypatch Tracer.tracer with a property to return a shim instance instead. This
    # makes all instances of Tracer (even those already created) use a ShimTracer.
    Tracer.tracer = property(fget_tracer, fset_tracer)
    _logger.info("Installed OpenCensus shim")


def uninstall_shim() -> None:
    if hasattr(Tracer, "tracer"):
        del Tracer.tracer



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The OpenTelemetry OpenCensus shim is a library which allows an easy migration from OpenCensus
to OpenTelemetry. Additional details can be found `in the specification
<https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/compatibility/opencensus.md>`_.

The shim consists of a set of classes which implement the OpenCensus Python API while using
OpenTelemetry constructs behind the scenes. Its purpose is to allow applications which are
already instrumented using OpenCensus to start using OpenTelemetry with minimal effort, without
having to rewrite large portions of the codebase.
"""

from opentelemetry.shim.opencensus._patch import install_shim, uninstall_shim

__all__ = [
    "install_shim",
    "uninstall_shim",
]

# TODO: Decide when this should be called.
# 1. defensive import in opentelemetry-api
# 2. defensive import directly in OpenCensus, although that would require a release
# 3. ask the user to do it
# install_shim()
